/*
 * Copyright E. Bigeon (2014)
 *
 * emmanuel@bigeon.fr
 *
 * This software is a computer program whose purpose is to
 * Socket implementation of GCLC.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
package fr.bigeon.gclc.socket;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.nio.charset.Charset;
import java.util.logging.Level;
import java.util.logging.Logger;

import fr.bigeon.gclc.ConsoleApplication;
import fr.bigeon.gclc.manager.ConsoleManager;
import fr.bigeon.gclc.manager.PipedConsoleManager;
import fr.bigeon.gclc.manager.ReadingRunnable;

/** This is a socket communicating console consoleManager
 * <p>
 * To use this application, the following flow should be used:
 *
 * <pre>
 * SocketConsoleApplicationShell shell = new SocketConsoleApplicationShell();
 * ConsoleApplication myApplication = new MyConsoleApplication(shell.getConsoleManager(), ...);
 * shell.setApplication(myApplication);
 * Thread th = new Thread(shell);
 * th.start();
 * </pre>
 * <p>
 * This will start the application in a separate thread. The application will be
 * listening on the given socket and writing back on it. If this is all your
 * application, you should then {@link Thread#join()} the thread to wait for the
 * end of the execution.
 *
 * @author Emmanuel Bigeon */
public class SocketConsoleApplicationShell implements Runnable, AutoCloseable {

    /** The runnable to forward output of application to socket.
     *
     * @author Emmanuel Bigeon */
    private final class OutputForwardRunnable implements Runnable {
        /**
         * 
         */
        private final PrintWriter writer;
        /**
         * 
         */
        private final Socket socket;

        /** @param writer the writer
         * @param socket the socket */
        protected OutputForwardRunnable(PrintWriter writer, Socket socket) {
            this.writer = writer;
            this.socket = socket;
        }

        @SuppressWarnings("synthetic-access")
        @Override
        public void run() {
            try {
                while (!socket.isClosed()) {
                    while (!socket.isClosed() && !consoleManager.available()) {
                        waitASec();
                    }
                    if (socket.isClosed()) {
                        return;
                    }
                    String m = consoleManager.readNextLine();
                    writer.println(m);
                }
            } catch (IOException e) {
                LOGGER.log(Level.SEVERE, "Unexpected problem in manager", //$NON-NLS-1$
                        e);
            }
        }

    }

    /** The end of line character */
    protected static final String EOL = "\n"; //$NON-NLS-1$
    /** The class logger */
    private static final Logger LOGGER = Logger
            .getLogger(SocketConsoleApplicationShell.class.getName());
    /** Time of wait */
    protected static final long ONE_TENTH_OF_SECOND = 100;
    /** The listening port */
    private final int port;
    /** The application */
    private ConsoleApplication app;
    /** The session closing command */
    private final String close;
    /** The running status */
    private boolean running;

    /** The console manager implementation */
    private final PipedConsoleManager consoleManager;
    /** The auto close flag. if this is true, every request closes the session
     * after its call */
    private final boolean autoClose;
    /** The server socket */
    private ServerSocket serverSocket;
    /** The application shutdown string */
    private final String applicationShutdown;
    /** The charset for the communication. */
    private final Charset charset;

    /** Create a socket application shell which will listen on the given port
     * and close session upon the provided string reception by client
     *
     * @param port the port to listen to
     * @param close the session closing command
     * @param applicationShutdown the appication shut down command
     * @param charset the charset for communication
     * @throws IOException if the manager could not be created */
    public SocketConsoleApplicationShell(int port, String close,
            String applicationShutdown, Charset charset) throws IOException {
        this.port = port;
        this.close = close;
        this.applicationShutdown = applicationShutdown;
        this.autoClose = false;
        this.charset = charset;
        //
        consoleManager = new PipedConsoleManager();
    }

    /** Create a socket application shell which will listen on the given port
     * and auto close session after one instruction
     *
     * @param port the port to listen to
     * @param autoClose if the session must be closed once the request has been
     *            sent
     * @param applicationShutdown the appication shut down command
     * @param charset the charset for communication
     * @throws IOException if the manager could not be created */
    public SocketConsoleApplicationShell(int port, boolean autoClose,
            String applicationShutdown, Charset charset) throws IOException {
        this.port = port;
        this.autoClose = autoClose;
        this.applicationShutdown = applicationShutdown;
        this.close = autoClose ? null : "close"; //$NON-NLS-1$
        this.charset = charset;
        //
        consoleManager = new PipedConsoleManager();
    }

    /* (non-Javadoc)
     * @see java.lang.Runnable#run() */
    @Override
    public void run() {
        // Create the server
        try (ServerSocket actualServerSocket = new ServerSocket(port)) {
            this.serverSocket = actualServerSocket;
            running = true;
            // Create the streams
            runSokectServer();
        } catch (final IOException e) {
            LOGGER.severe("Communication error between client and server"); //$NON-NLS-1$
            LOGGER.log(Level.FINE,
                    "Communication error between client and server", e); //$NON-NLS-1$
        }
    }

    /** @throws IOException if the communication with the client failed */
    private void runSokectServer() throws IOException {
        final ConsoleRunnable runnable = new ConsoleRunnable(app);
        Thread appThNext = new Thread(runnable, "gclc-ctrl"); //$NON-NLS-1$
        appThNext.start();
        while (running) {
            LOGGER.info("Waiting client"); //$NON-NLS-1$
            try (Socket clientSocket = serverSocket.accept();
                 PrintWriter out = new PrintWriter(new OutputStreamWriter(
                         clientSocket.getOutputStream(), charset), true);
                 InputStreamReader isr = new InputStreamReader(
                         clientSocket.getInputStream(), charset);
                 BufferedReader in = new BufferedReader(isr);) {
                // this is not threaded to avoid several clients at the same
                // time
                LOGGER.info("Opening client"); //$NON-NLS-1$

                // Initiate application
                if (!runnable.isApplicationRunning()) {
                    LOGGER.info("Start application"); //$NON-NLS-1$
                    startApplication(runnable);
                } else {
                    LOGGER.info("Reconnect to application"); //$NON-NLS-1$
                    out.println("Reconnected"); //$NON-NLS-1$
                    out.println(consoleManager.getPrompt());
                }
                communicate(clientSocket, out, in);
            } catch (SocketException e) {
                LOGGER.log(Level.INFO, "Socket closed"); //$NON-NLS-1$
                LOGGER.log(Level.FINE,
                        "Socket closed with exception (probably due to server interruption)", //$NON-NLS-1$
                        e);
            } catch (IOException e) {
                throw e;
            }
            LOGGER.info("Closing client"); //$NON-NLS-1$
        }
        runnable.setRunning(false);
        try {
            consoleManager.type(applicationShutdown);
            consoleManager.close();
        } catch (IOException e) {
            LOGGER.warning("Unable to close application correctly"); //$NON-NLS-1$
            LOGGER.log(Level.FINE, "Application closing caused an exception", //$NON-NLS-1$
                    e);
        }
        LOGGER.info("Closing Server"); //$NON-NLS-1$
    }

    /** @param runnable the runnable */
    private void startApplication(ConsoleRunnable runnable) {
        runnable.restart();
        synchronized (this) {
            try {
                wait(ONE_TENTH_OF_SECOND);
            } catch (InterruptedException e) {
                LOGGER.log(Level.SEVERE, "Interruption in application start", //$NON-NLS-1$
                        e);
            }
        }
    }

    /** active communication between server and client
     * 
     * @param socket the socket
     * @param writer the writer to the application
     * @param in the input from the client
     * @throws IOException if the communication failed */
    private void communicate(final Socket socket, final PrintWriter writer,
                             BufferedReader in) throws IOException {
        OutputForwardRunnable cc = new OutputForwardRunnable(writer, socket);
        Thread th = new Thread(cc, "ClientComm"); //$NON-NLS-1$
        th.start();
        if (autoClose) {
            communicateOnce(in);
        } else {
            communicateLoop(in);
        }
    }

    /** @param in the input from the client
     * @throws IOException if the communication failed */
    private void communicateOnce(BufferedReader in) throws IOException {
        ReadingRunnable reading = new ReadingRunnable(in);
        Thread th = new Thread(reading, "gclcToApp"); //$NON-NLS-1$
        th.start();
        communicationContent(reading);
        doEndCommunication(reading);
    }

    /** @param reading the reading runnable
     * @throws IOException if the end of communication failed */
    private void doEndCommunication(ReadingRunnable reading) throws IOException {
        reading.setRunning(false);
        Thread wait = consoleManager.getWaitForDelivery("Bye."); //$NON-NLS-1$
        consoleManager.println("Bye."); //$NON-NLS-1$
        try {
            wait.join();
        } catch (InterruptedException e) {
            LOGGER.warning("The Bye wait was interrupted."); //$NON-NLS-1$
            LOGGER.log(Level.FINE, "An interruption occured", e); //$NON-NLS-1$
        }
    }

    /** @param in the input from the client
     * @throws IOException if the communication failed */
    private void communicateLoop(BufferedReader in) throws IOException {
        ReadingRunnable reading = new ReadingRunnable(in);
        Thread th = new Thread(reading, "gclcToApp"); //$NON-NLS-1$
        th.start();
        while (app.isRunning() && communicationContent(reading)) {
            // keep on going
        }
        doEndCommunication(reading);
    }

    /** @param reading the reading
     * @return if the communication should be stopped.
     * @throws IOException if the reading failed */
    private boolean communicationContent(ReadingRunnable reading) throws IOException {
        try {
            while (app.isRunning() && !reading.hasMessage()) {
                synchronized (this) {
                    waitASec();
                }
            }
        } catch (IOException e) {
            LOGGER.warning("Client seems dead. Closing communication"); //$NON-NLS-1$
            LOGGER.log(Level.FINE, "Wait on message from client failed", e); //$NON-NLS-1$
            return false;
        }
        if (!app.isRunning()) {
            return false;
        }
        String ln = reading.getMessage();
        if (ln.equals(close)) {
            return false;
        }
        // Pass command to application
        consoleManager.type(ln);
        return true;
    }

    /** @return the consoleManager */
    public synchronized ConsoleManager getConsoleManager() {
        return consoleManager;
    }

    /** @param app the application to set */
    public synchronized void setApplication(ConsoleApplication app) {
        this.app = app;
    }

    /** This method will request the server to stop.
     * <p>
     * In most cases, this will terminate communication on every client. On some
     * cases, the */
    public void stop() {
        running = false;
        try {
            serverSocket.close();
        } catch (IOException e) {
            LOGGER.log(Level.SEVERE, "Exception in closing socket server", e); //$NON-NLS-1$
        }
        app.exit();
    }

    /** a method to wait some time */
    protected void waitASec() {
        try {
            synchronized (this) {
                wait(ONE_TENTH_OF_SECOND);
            }
        } catch (InterruptedException e) {
            LOGGER.log(Level.SEVERE, "Interrupted wait", //$NON-NLS-1$
                    e);
            return;
        }
    }

    /* (non-Javadoc)
     * @see java.lang.AutoCloseable#close() */
    @Override
    public void close() throws IOException {
        consoleManager.close();
    }
}
